import json

from xml.etree.ElementTree import Element
from xml.etree import cElementTree as eTree

from messagebus import (
    ApplicationErrorMessage,
    MessageBus,
    Message,
    DataType,
    DefaultToken,
)
from pos_api import PosService
from product_api import ProductService
from typing import (
    Dict,
    Optional,
    Union,
    List,
)
from .dto import DeliveryInfoDto
from .dto.exception import (
    OrderException,
    ProductInRupture,
    UnavailableProductStock,
)

from .error import OrderError
from .token import OrderApiTokens


class OrderService(object):

    def __init__(
        self,                   # type: OrderService
        message_bus,            # type: MessageBus
        pos_id,                 # type: int
        pos_service=None,       # type: Optional[PosService]
        product_service=None,   # type: Optional[ProductService]
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.pos_id = pos_id

        self.component_name = "ORDERMGR{}".format(self.pos_id)
        if pos_service is not None:
            self.pos_service = pos_service
        else:
            self.pos_service = PosService(
                message_bus=message_bus,
                pos_id=pos_id,
            )

        if product_service is not None:
            self.product_service = product_service
        else:
            self.product_service = ProductService(message_bus=message_bus)

    def create_order(
        self,                       # type: OrderService
        sale_type,                  # type: str
        session_id=None,            # type: Optional[str]
        pod_type=None,              # type: Optional[str]
        order_type="SALE",          # type: str
        order_sub_type=None,        # type: Optional[str]
        service_period_id=None,     # type: Optional[int]
        delivery_info=None,         # type: Optional[DeliveryInfoDto]
        additional_info=None,       # type: Optional[str]
        custom_properties=None,     # type: Optional[Dict[str, str]]
        blk_notify=False,           # type: bool
    ):
        # type: (...) -> Element

        if session_id is None:
            session_id = self.pos_service.get_operator_session()

        if pod_type is None:
            pod_type = self.pos_service.get_pod_type()

        request = {
            "blkNotify": blk_notify,
            "posId": self.pos_id,
            "sessionId": session_id,
            "saleType": sale_type,
            "podType": pod_type,
            "orderType": order_type,
        }
        if order_sub_type is not None:
            request["orderSubType"] = order_sub_type
        if additional_info is not None:
            request["additionalInfo"] = additional_info
        if service_period_id is not None:
            request["servicePeriodId"] = service_period_id
        if custom_properties is not None:
            request["customProperties"] = custom_properties
        if delivery_info is not None:
            request["deliveryInfo"] = delivery_info.to_dict()

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=OrderApiTokens.TK_SLCTRL_OMGR_CREATEORDER,
                data_type=DataType.param,
                data=json.dumps(request).encode("utf-8"),
                timeout=600000000,
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=msg)

        return eTree.XML(msg.data)

    def do_sale(
        self,                               # type: OrderService
        item_id,                            # type: str
        qty=1,                              # type: float
        apply_default_options=True,         # type: bool
        custom_properties=None,             # type: Dict[str, str]
        blk_notify=False,                   # type: bool
    ):
        # type: (...) -> eTree.Element

        if custom_properties is not None:
            custom_properties = json.dumps(custom_properties)

        params = (
            "1" if blk_notify else "0",
            self.pos_id,
            item_id,
            qty,
            "1" if apply_default_options else "0",
            custom_properties or "",
        )
        data = "\x00".join((str(p) for p in params))
        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_DOSALE,
            data=data,
        )

    def do_option(
        self,                           # type: OrderService
        item_id,                        # type: str
        qty,                            # type: Union[str, int]
        line_number="",                 # type: Union[str, int]
        apply_default_options=False,    # type: bool
        blk_notify=False,               # type: bool
    ):
        # type: (...) -> eTree.Element

        blk_notify = "1" if blk_notify else "0"
        apply_default_options = "1" if apply_default_options else "0"
        params = (blk_notify, self.pos_id, item_id, qty, line_number, "", apply_default_options)
        buffer = "\0".join(str(p) for p in params)

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_DOOPTION,
            data=buffer,
        )

    def do_tender(
        self,                   # type: OrderService
        tender_id,              # type: Union[int, str]
        amount,                 # type: Union[float, str]
        do_not_close=False,     # type: bool
        order_tender_id=None,   # type: Optional[Union[int, str]]
        tip_amount=None,        # type: Optional[float]
        tender_detail=None,     # type: Optional[Union[Dict[str, Union[bool, int, str]], str]]
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "tenderId": int(tender_id),
            "amount": float(amount),
            "doNotClose": do_not_close,
            "orderTenderId": int(order_tender_id) if order_tender_id not in [None, ""] else None,
            "tipAmount": float(tip_amount) if tip_amount not in [None, ""] else None,
            "tenderDetail": json.dumps(tender_detail) if isinstance(tender_detail, dict) else tender_detail,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_DOTENDER,
            data=json.dumps(request),
        )

    def change_modifier(
        self,               # type: OrderService
        line_number,        # type: int
        item_id,            # type: str
        level,              # type: int
        part_code,          # type: str
        new_qty,            # type: Union[str, int]
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        blk_notify = "1" if blk_notify else "0"
        params = (blk_notify, self.pos_id, line_number, item_id, level, part_code, new_qty)
        buffer = "\0".join(str(p) for p in params)

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_CHANGEMODIFIER,
            data=buffer,
        )

    def clear_option(
        self,               # type: OrderService
        line_number,        # type: Union[int, str]
        quantity="",        # type: Union[int, str]
        item_id="",         # type: str
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        params = ("1" if blk_notify else "0", self.pos_id, line_number, item_id, quantity)

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_CLEAROPTION,
            data="\x00".join((str(p) for p in params)),
        )

    def get_order_picture(
        self,               # type: OrderService
        order_id=None,      # type: int
        no_cache=False,     # type: bool
    ):
        # type: (...) -> Optional[Element]

        no_cache = "true" if no_cache else "false"
        data = "{}\0{}\0{}".format(str(self.pos_id), str(order_id or ""), no_cache).encode("utf-8")
        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=OrderApiTokens.TK_SLCTRL_OMGR_ORDERPICT,
                data_type=DataType.param,
                data=data,
                timeout=600000000,
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=msg)

        order_str = msg.data.split(b"\0")[2]
        if not order_str:
            return None

        order_picture = eTree.XML(order_str)
        if order_id and order_picture.tag != "Order":
            order_picture = order_picture.findall(".//Order")[0]

        return order_picture

    def has_current_order(
        self,       # type: OrderService
    ):
        # type: (...) -> bool

        order_picture = self.get_order_picture()
        if order_picture is None:
            return False

        state = order_picture.get("state")
        return state in ["IN_PROGRESS", "TOTALED"]

    def void_lines(
        self,                   # type: OrderService
        line_numbers,           # type: List[int]
        blk_notify=False,       # type: bool
        custom_property=None,   # type: Optional[List[Dict[str, str]]]
    ):
        # type: (...) -> None

        if custom_property:
            custom_properties = json.dumps(custom_property)
        else:
            custom_properties = "[]"

        line_numbers_str = "|".join(map(str, line_numbers))
        params = ("1" if blk_notify else "0", self.pos_id, line_numbers_str, custom_properties)

        msg = self._get_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_VOIDLINE,
            data="\x00".join((str(p) for p in params)),
        )

        return eTree.XML(msg.data)

    def void_order(
        self,                   # type: OrderService
        order_id="",            # type: Union[str, int]
        abandon=False,          # type: bool
        order_status="",        # type: str
        void_reason="",         # type: str
        blk_notify=False,       # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "orderId": int(order_id) if order_id else None,
            "abandon": abandon,
            "orderStatus": int(order_status) if order_status not in ("", None) else None,
            "voidReason": void_reason if void_reason else None,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_VOIDORDER,
            data=json.dumps(request),
        )

    def store_order(
        self,               # type: OrderService
        blk_notify=False,   # type: str
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_STOREORDER,
            data=json.dumps(request),
        )

    def recall_order(
        self,                       # type: OrderService
        order_id,                   # type: Union[int, str]
        session_id="",              # type: str
        blk_notify=False,           # type: str
        notify_ndiscount=True,      # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "orderId": int(order_id),
            "sessionId": session_id,
            "blkNotify": blk_notify,
            "notifyNdiscount": notify_ndiscount,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_RECALLORDER,
            data=json.dumps(request),
        )

    def pull_orders(
        self,           # type: OrderService
        orders_ids,     # type: List[int]
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "ordersIds": orders_ids,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_PULLORDERS,
            data=json.dumps(request),
        )

    def reopen_order(
        self,               # type: OrderService
        blk_notify=False,   # type: Optional[bool]
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_ORDERREOPEN,
            data=json.dumps(request),
        )

    def total_order(
        self,                   # type: OrderService
        blk_notify=False,       # type: bool
        notify_ndiscount=True,  # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "notifyNdiscount": notify_ndiscount,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_ORDERTOTAL,
            data=json.dumps(request),
        )

    def change_line_quantity(
        self,               # type: OrderService
        line_number,        # type: int
        item_id,            # type: str
        new_quantity,       # type: int
        blk_notify=False,   # type: bool
    ):
        # type: (...) -> eTree.Element

        request = {
            "posId": int(self.pos_id),
            "blkNotify": blk_notify,
            "lineNumber": line_number,
            "itemId": item_id,
            "newQuantity": new_quantity,
        }

        return self._get_order_picture_from_success_message_response(
            token=OrderApiTokens.TK_SLCTRL_OMGR_CHANGEQTY,
            data=json.dumps(request),
        )

    def list_orders(
        self,                  # type: OrderService
        state="",              # type: Optional[str]
        originator_id="",      # type: Optional[str]
        pod_type="",           # type: Optional[str]
        since="",              # type: Optional[str]
        until="",              # type: Optional[str]
        session_id="",         # type: Optional[str]
        period="",             # type: Optional[str]
        limit="",              # type: Optional[str]
        descending="",         # type: Optional[str]
        instance_number="",    # type: Optional[str]
        include_property="",   # type: Optional[str]
        exclude_property="",   # type: Optional[str]
        property_val="",       # type: Optional[str]
        order_by="",           # type: Optional[str]
        show_tenders=False,    # type: Optional[bool]
    ):
        # type: (...) -> List[Dict[str, Union[str, dict, list]]]

        params = (
            state,
            originator_id,
            pod_type,
            since,
            until,
            session_id,
            period,
            limit,
            include_property,
            exclude_property,
            property_val,
            order_by,
            "1" if descending and descending != "0" else "",
            instance_number,
            "1" if show_tenders is True else "0",
        )

        data = "\0".join((str(p) for p in params)).encode("utf-8")
        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=OrderApiTokens.TK_SLCTRL_OMGR_LISTORDERS,
                data_type=DataType.param,
                data=data,
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=msg)

        return self._parse_listed_orders(orders_data=msg.data.decode("utf-8"))

    @staticmethod
    def _parse_listed_orders(
        orders_data,    # type: str
    ):
        # type: (...) -> List[Dict[str, Union[str, dict, list]]]

        response_params = orders_data.split("\0")
        xml = eTree.XML(response_params[2])
        result = []
        for order in xml.findall("Order"):
            order_dict = dict(order.items())
            order_dict["custom_properties"] = {}
            for order_property in order.findall("CustomOrderProperties/OrderProperty"):
                order_dict["custom_properties"][order_property.get("key")] = order_property.get("value")

            order_dict["tender_history"] = []
            for tender in order.findall("TenderHistory/Tender"):
                order_dict["tender_history"].append(dict(tender.items()))

            result.append(order_dict)

        return result

    @staticmethod
    def _build_order_exception(
        msg,  # type: Message
    ):
        # type: (...) -> OrderException

        try:
            data_text = msg.data.decode("utf-8")
        except UnicodeDecodeError:
            data_text = msg.data.hex()

        parts = data_text.split("\0")
        if len(parts) >= 2:
            error_code = parts[0]
            error_message = parts[1]
            return OrderException(
                code=error_code,
                description=error_message,
            )

        return OrderException(
            code=-1,
            description="Unknown error: {}".format(data_text),
        )

    def _get_order_picture_from_success_message_response(
        self,   # type: OrderService
        token,  # type: int
        data,   # type: str
    ):
        # type: (...) -> eTree.Element

        reply = self._get_success_message_response(
            token=token,
            data=data,
        )

        return eTree.XML(reply.data)

    def _get_success_message_response(
        self,   # type: OrderService
        token,  # type: int
        data,   # type: str
    ):
        # type: (...) -> Message

        reply = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=token,
                data_type=DataType.param,
                data=data.encode("utf-8"),
            ),
        )

        if reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            application_error = ApplicationErrorMessage.build_from_message(message=reply)
            if application_error.error_code == OrderError.PRODUCT_IN_RUPTURE.code:
                raise ProductInRupture(
                    part_code=int(application_error.error_message),
                    part_name=self._get_part_name(application_error.error_message),
                )
            if application_error.error_code == OrderError.UNAVAILABLE_PRODUCT_STOCK.code:
                raise UnavailableProductStock(
                    part_code=int(application_error.error_message),
                    part_name=self._get_part_name(application_error.error_message),
                )

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise self._build_order_exception(msg=reply)

        return reply

    def _get_part_name(
        self,       # type: OrderService
        part_code,  # type: str
    ):
        # type: (...) -> str

        products_name = self.product_service.get_products_name()

        return products_name.get(part_code, part_code)



from api.order import get_order_service
from application.domain import OrderManager, SaleLine  # noqa
from sysactions import get_posot, get_model


class MsgBusOrderManager(OrderManager):

    def recall_order(self, pos_id, order_id):
        order_service = get_order_service(pos_id=pos_id)
        order_service.recall_order(order_id=order_id)

    def store_order(self, pos_id):
        order_service = get_order_service(pos_id=pos_id)
        order_service.store_order()

    def change_line_seat(self, pos_id, sale_line, seat_number):
        # type: (str, SaleLine, str) -> None

        model = get_model(pos_id)
        pos_ot = get_posot(model)

        pos_ot.setOrderCustomProperty(
            "seat",
            str(seat_number),
            sale_line.order_id,
            sale_line.line_number,
            sale_line.item_id,
            sale_line.level,
            sale_line.part_code
        )

    def get_current_order_state(self, pos_id):
        order_service = get_order_service(pos_id=pos_id)

        order_picture = order_service.get_order_picture()
        if order_picture is not None:
            return order_picture.get("state")
        return None



# -*- coding: utf-8 -*-
import json
from xml.etree import cElementTree as eTree

from typing import (
    Dict,
    Optional,
)

from actions.delivery import is_delivery_fee_enabled
from actions.delivery.get_delivery_fee_code import get_delivery_fee_code
from actions.functions.do_store_order import doStoreOrder
from api.order import get_order_service
from old_helper import round_half_away_from_zero
from sysactions import action, get_model, get_posot, get_operator_session, StopAction, translate_message, \
    show_confirmation, get_current_order

from bustoken import TK_REMOTE_ORDER_SEND_ORDER_TO_PRODUCTION
from msgbus import FM_PARAM, TK_SYS_ACK
from sysactions import show_messagebox

from mw_helper import format_currency

from custom_sysactions import (
    block_action_if_order_items_not_resolved,
    fix_current_order_in_model,
    validate_order_state,
)
from model.customexception import (
    OrderNotFoundException,
    SendOrderToProductionError,
)
from .. import logger
from .. import mb_context
from modelcommon.remoteorder import DeliveryDiscounts


@action
def produce_delivery_order(
    pos_id,                         # type: str
    order_id,                       # type: str
    warn_inconsistencies='false',   # type: Optional[str]
):
    # type: (...) -> None

    produce_order(
        pos_id=pos_id,
        order_id=order_id,
        warn_inconsistencies=warn_inconsistencies,
    )


@action
@validate_order_state
@block_action_if_order_items_not_resolved
def produce_current_delivery_order(
    pos_id,                         # type: str
    warn_inconsistencies='false',   # type: Optional[str]
):
    # type: (...) -> None

    produce_order(
        pos_id=pos_id,
        order_id=None,
        warn_inconsistencies=warn_inconsistencies,
    )


def produce_order(
    pos_id,                         # type: str
    order_id=None,                  # type: Optional[str]
    warn_inconsistencies='false',   # type: Optional[str]
):
    # type: (...) -> None

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    session = get_operator_session(model)

    order_service = get_order_service(pos_id=pos_id)
    if order_id is None:
        order_in_current_pos = True
        model_order = model.find(".//Order")
        order_id = model_order.get("orderId")
    else:
        order_in_current_pos = False
        order_service.recall_order(order_id=order_id)

    pos_ot.clearDiscount(1)

    if not _validate_manual_order(model, pos_ot, pos_id, order_id, warn_inconsistencies.lower() != 'false'):
        return

    try:
        try:
            doStoreOrder(
                pos_id=pos_id,
                show_popup=False,
            )
        except OrderNotFoundException as ex:
            fixed = fix_current_order_in_model(
                pos_id=pos_id,
                model=model,
                order_taker=pos_ot,
                order_id=order_id,
            )
            if not fixed:
                raise ex

        send_remote_order_to_production(order_id=order_id)
        show_messagebox(pos_id, "$ORDER_PRODUCED")
    except Exception as ex:
        logger.exception("Error producing delivery order")
        show_messagebox(pos_id, message=ex.message)

        if order_in_current_pos:
            order_service.recall_order(
                order_id=order_id,
                session_id=session,
            )

        raise StopAction()


def send_remote_order_to_production(
    order_id,       # type: str
):
    # type: (...) -> None

    msg = mb_context.MB_EasySendMessage(
        "RemoteOrder",
        TK_REMOTE_ORDER_SEND_ORDER_TO_PRODUCTION,
        FM_PARAM,
        order_id
    )
    if msg.token != TK_SYS_ACK:
        raise SendOrderToProductionError("$ERROR_SENDING_REMOTE_ORDER_TO_PRODUCTION|{}".format(msg.data))


def _validate_manual_order(model, pos_ot, pos_id, order_id, warn_inconsistencies):
    try:
        order_picture = eTree.XML(pos_ot.orderPicture(pos_id, order_id))
        manual_order = order_picture.find("Order") if order_picture.find("Order") else order_picture
        remote_order_json_element = order_picture.find(".//OrderProperty[@key='REMOTE_ORDER_JSON']")
        remote_order_json_text = None
        if remote_order_json_element is not None:
            remote_order_json_text = remote_order_json_element.get("value", None)
        if not remote_order_json_text or not warn_inconsistencies:
            return True

        remote_json = json.loads(remote_order_json_text)

        order_items_quantity = _get_order_items_quantity(manual_order)
        order_total_value = _get_order_total_value(manual_order)
        json_items_quantity = _get_json_items_quantity(remote_json)
        json_total_value = _get_json_total_value(remote_json)

        quantity_is_different, quantity_diff = _items_quantity_is_different(order_items_quantity, json_items_quantity)
        value_is_different, value_diff = _value_is_different(order_total_value, json_total_value)

        if quantity_is_different or value_is_different:
            _confirm_order_difference(model, pos_id, remote_json)
            if quantity_is_different:
                _validate_quantity_difference(model, pos_id, order_items_quantity, json_items_quantity)
            if value_is_different:
                _validate_value_difference(model, pos_id, value_diff, order_total_value, json_total_value)

    except StopAction:
        return False
    except (Exception,):
        logger.exception("Error validating manual order")
        return False

    return True


def _validate_value_difference(model, pos_id, value_diff, order_total_value, json_total_value):
    message = "$CONFIRM_MANUAL_ORDER_DISCOUNT" if value_diff > 0 else "$CONFIRM_MANUAL_ORDER_ADDITION"
    message = message + "|{0}|{1}|{2}".format(
        format_currency(
            model=model,
            value=order_total_value,
        ),
        format_currency(
            model=model,
            value=json_total_value,
        ),
        format_currency(
            model=model,
            value=abs(value_diff),
        )
    )
    if not show_confirmation(pos_id, message):
        raise StopAction()

    _apply_difference_discount(pos_id, value_diff)


def _validate_quantity_difference(model, pos_id, order_items_quantity, json_items_quantity):
    message = "$CONFIRM_MANUAL_ORDER_ITEM_QUANTITY_DIFF"
    message = message + "|{0}|{1}".format(order_items_quantity, json_items_quantity)
    if not show_confirmation(pos_id, message):
        raise StopAction()


def _apply_difference_discount(pos_id, value_diff):
    from actions.discount.apply_discount import apply_amount_discount

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order = get_current_order(model)
    apply_amount_discount(pos_id, pos_ot, order, discount_type=1, discount_value=float(value_diff))


def _confirm_order_difference(model, pos_id, remote_json):
    partner_order_id = _get_json_order_id(remote_json)
    message = translate_message(model, "DELIVERY_RECALL_WARNING")
    message = message.format(
        partner_order_id,
        remote_json.get("partner", "").upper(),
        str(_get_json_items_quantity(remote_json)),
        format_currency(
            model=model,
            value=_get_json_total_value(remote_json),
        )
    )
    if not show_confirmation(pos_id, message):
        raise StopAction()


def _items_quantity_is_different(order_items_quantity, json_items_quantity):
    is_different = not (order_items_quantity == json_items_quantity)
    value_diff = order_items_quantity - json_items_quantity

    return is_different, value_diff


def _value_is_different(order_total_value, json_total_value):
    round_order_value = round_half_away_from_zero(order_total_value, 2)
    round_json_value = round_half_away_from_zero(json_total_value, 2)
    is_different = not (round_order_value == round_json_value)
    value_diff = round_half_away_from_zero(order_total_value - json_total_value, 2)

    return is_different, value_diff


def _get_order_total_value(manual_order):
    manual_order_total = float(manual_order.get("totalAfterDiscount", 0)) + float(manual_order.get("taxTotal", 0))
    return manual_order_total


def _get_json_total_value(
    remote_json,    # type:Dict
):
    # type: (...) -> float

    vouchers = remote_json.get("vouchers")
    delivery_discounts = DeliveryDiscounts(logger=logger)
    json_discounts = delivery_discounts.get_discount_amount_by_type(
        vouchers=vouchers,
        discount_type="store",
    )
    json_total_value = float(remote_json.get("subTotal", 0)) - json_discounts

    delivery_fee_value = 0
    if is_delivery_fee_enabled():
        json_delivery_fee_discount = delivery_discounts.get_delivery_fee_discounts(vouchers=vouchers)
        delivery_fee_value = float(remote_json.get("deliveryFee", 0)) - json_delivery_fee_discount

    return json_total_value + delivery_fee_value


def _get_order_items_quantity(manual_order):
    sale_lines = manual_order.findall(".//SaleLine[@level='0']")
    delivery_fee_code = get_delivery_fee_code()

    return sum(float(c.get("qty", "0")) for c in _get_valid_sale_lines(sale_lines, delivery_fee_code))


def _get_valid_sale_lines(sale_lines, fee_code):
    return [line for line in sale_lines if (float(line.get("qty", "0")) > 0 and line.get("partCode", "") != fee_code)]


def _get_json_items_quantity(remote_json):
    valid_items = [line for line in remote_json.get("items", []) if (float(line.get("quantity", "0")) > 0)]

    return sum(float(c.get("quantity", "0")) for c in valid_items)


def _get_json_order_id(remote_json):
    return remote_json.get("shortReference", "")



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from concurrent.futures import CancelledError


class RetriesExceededError(Exception):
    def __init__(self, last_exception, msg='Max Retries Exceeded'):
        super().__init__(msg)
        self.last_exception = last_exception


class S3UploadFailedError(Exception):
    pass


class InvalidSubscriberMethodError(Exception):
    pass


class TransferNotDoneError(Exception):
    pass


class FatalError(CancelledError):
    """A CancelledError raised from an error in the TransferManager"""

    pass
